// ----------------------------------------------------------------------------
// time functions

#pragma GCC diagnostic ignored "-Wwrite-strings"            // always ignore warning about using strings as constants
//#pragma GCC diagnostic ignored "-Wunused-variable"        // ignore unused variables until near end of development cycle, then comment this

#include                <Arduino.h>
#include                "globals.h"
#include                <time.h>                            // for ctime

#include                <DS1307RTC.h>

// ---------------------------------------------------------------------------
// Public Globals
// external's in h file defined here
// ---------------------------------------------------------------------------

time_t                  ti_bootTime     = 0;                // unix timestamp when first valid time is obtained
bool                    ti_DLS;                             // are we currently observing DLS time?
bool                    ti_valid        = false;            // is the RTC valid?
const char *            ti_zones[ti_zonesMax] = {           // TZ names corr to myTz 0-3
                                            "East",
                                            "Central",
                                            "Mountain",
                                            "Pacific",
                                            };

// ---------------------------------------------------------------------------
// Private Globals
// ---------------------------------------------------------------------------

int                     myTz;                               // timezone we are in

// These are the supported timezone rules with normal & daylight savings
TimeChangeRule          EDT             = {"EDT", Second, Sun, Mar, 2, -4*60};
TimeChangeRule          EST             = {"EST", First,  Sun, Nov, 2, -5*60};
TimeChangeRule          CDT             = {"CDT", Second, Sun, Mar, 2, -5*60};
TimeChangeRule          CST             = {"CST", First,  Sun, Nov, 2, -6*60};
TimeChangeRule          MDT             = {"MDT", Second, Sun, Mar, 2, -6*60};
TimeChangeRule          MST             = {"MST", First,  Sun, Nov, 2, -7*60};
TimeChangeRule          PDT             = {"PDT", Second, Sun, Mar, 2, -7*60};
TimeChangeRule          PST             = {"PST", First,  Sun, Nov, 2, -8*60};

// These are the tzRules above, encoded into an array
TimeChangeRule          tzRules[4][2] = {
                            {EDT, EST},
                            {CDT, CST},
                            {MDT, MST},
                            {PDT, PST},
                            };

bool                    useDLS          = true;             // will we be observing DLS changes?

// ===========================================================================

// ---------------------------------------------------------------------------
// Keeps clock in sync with DS3231 RTC.

// This function is passed to setSyncProvider. It will intialize and maintain
// the MCU's clock.

time_t getDS3231Time(
    void
    ) {

return RTC.get();

} // getDS3231Time


// ---------------------------------------------------------------------------
// Keeps clock in sync with hardware Teensy RTC.

// This function is passed to setSyncProvider. It will intialize and maintain
// the MCU's clock.

time_t getTeensyTime(
    void
    ) {

return Teensy3Clock.get();

} // getTeensyTime

// ---------------------------------------------------------------------------
// This function is used by setSyncProvider when there is no RTC. It will initialize
// the clock at Jan 1, 2023 12:00AM. It doesn't update the clcok so timeStatus will
// end up returning timeNeessync.

// This function is passed to setSyncProvider. It will intialize and maintain
// the MCU's clock.


time_t internalTime(
    void
    ) {

static bool             firstTime           = true;

if (firstTime) {
    firstTime = false;
    return 1672531200;                                      // we have no RTC so initialize to 1/1/23 12:00AM
    }
else
    return 0;                                               // no RTC, so no further updates

} // internalTime

// ---------------------------------------------------------------------------
// returns date/time as mm/dd/yy hh:mm:ss

char * ti_getDateTime(
    char *              s,
    int                 maxlen
    ) {

if (maxlen < 20) {
    s[0] = '\0';
    return s;
    }

if (!ti_valid) {
    snprintf(s, maxlen,
        "00/00/00 00:00:00");
    return s;
    }

snprintf(s, maxlen,
    "%02d/%02d/%04d %02d:%02d:%02d",
    month(), day(), year(), hour(), minute(), second()
    );

return s;

} // ti_getDateTime

// ---------------------------------------------------------------------------
// return time as hh:mm xM

char * ti_getTime(
    char *              s,
    int                 maxlen
    ) {

if (maxlen < 9) {
    s[0] = '\0';
    return s;
    }

if (!ti_valid) {
    snprintf(s, maxlen,
        "00:00:00");
    return s;
    }

snprintf(s, maxlen,
    "%02d:%02d:%02d", hour(), minute(), second());

return s;

} // ti_getTime

// ---------------------------------------------------------------------------
// TZ # into name (e.g. PST)

char * ti_getTZ(
    char *              s,
    int                 smax
    ) {

 if (myTz >= 0 && myTz < ti_zonesMax)
    snprintf(s, smax,
        "%s", ti_zones[myTz]);
else
    strcpy(s, "");

return s;

} // ti_getTZ

// ---------------------------------------------------------------------------
// Initialize time module

void ti_init(
    int                 tz,
    bool                useDLSp
    ) {

char                    s[40];
time_t                  timeStamp;

// Setup 'tz' with correct timechangerule
Timezone                tzDls(tzRules[myTz][0],
                            tzRules[myTz][1]);
Timezone                tzNoDls(tzRules[myTz][1]);

if (debugMode)
    serdb.printf("@ti_init: tz: %d; useDLS: %d; ", tz, useDLSp);

useDLS = useDLSp;
if (tz >= 0 && tz <= 3)
    myTz = tz;
else {
    if (debugMode)
        serdb.printf("Invalid Timezone: %d. Set to 3. ", tz);
    myTz = 3;
    }

#if defined rtcTeensy
    // Teensy RTC is set to a valid time at compile time so no "initial" setting
    // is necessary
    setSyncProvider(getTeensyTime);                         // get time from the Teensy RTC
#elif defined rtcDS3231
    // observed the initial time of DS3231 (upon inserting battery) is jan 1 2000.
    // That 'bad' time will be trapped below and ti_valid will not get set
    setSyncProvider(getDS3231Time);
#else
    // there is no RTC to provide initial time or time adjustments
#endif

if (debugMode)
    serdb.printf("timeStatus: %d\n", timeStatus());       // enum {timeNotSet, timeNeedsSync, timeSet}

if(timeStatus()!= timeSet) {                                // did setting time fail?
    ti_valid = false;                                       // The time is unknown
    ti_DLS   = true;                                        // DLS also unknown, but set it to a fixed value
    if (debugMode)
        serdb.printf("Unable to initialize Time!!!\n");
    return;
    }

// Date/Time overrides for testing - uncomment correct t3c.set & setTime
// Teensy3Clock.set(1699149590);                            // November 5, 2023 1:59:50 AM
// Teensy3Clock.set(1710035990);                            // Mar 10, 2024 1:59:50 AM
// setTime(Teensy3Clock.get());

timeStamp = now();

// Are we currently observing DLS time?
if (useDLS)
    ti_DLS = tzDls.locIsDST(timeStamp);
else
    ti_DLS = false; // we never use DLS in this situation

if (timeStamp < 1672531200) {				                // Jan 1, 2023
    ti_valid = false;
    return;
    }

ti_bootTime = now();                                        // save boot time

ti_valid = true;

if (debugMode) {
    serdb.printf("  now(): %s; ti_DLS: %d\n",
    ti_getDateTime(s, sizeof(s)), ti_DLS);
    }

} // ti_init

// ---------------------------------------------------------------------------
// change the time zone and correct the clock

void ti_newTZ(
    int                 tz                                  // timezone to change to, 0-3: EST, CST, MST, PST
    ) {

time_t                  gmt;
time_t                  oldTime;
time_t                  newTime;
Timezone                tzDls(tzRules[myTz][0],
                            tzRules[myTz][1]);
Timezone                tzNoDls(tzRules[myTz][1]);

if (debugMode)
    serdb.printf("@ti_newTZ: tz: %d\n", tz);

if (tz == myTz)                                             // is there actually a change?
    return;

oldTime = now();
if (useDLS)
    gmt = tzDls.toUTC(oldTime);
else
    gmt = tzNoDls.toUTC(oldTime);

myTz = tz;
ti_setRTC(gmt);

newTime = now();

if (debugMode) {
    serdb.printf("  old tstamp: %s", ctime(&oldTime));
    serdb.printf("  GMT tstamp: %s", ctime(&gmt));
    serdb.printf("  new tstamp: %s", ctime(&newTime));
    }

} // ti_newTZ

// ---------------------------------------------------------------------------
// If ti_init was already called and ti_valid is not set, when we finally get
// a good time set into RTC, we call ti_reinit to reset the time module.

void ti_reinit(                                             // used when NOT ti_valid and now we have a good time
    ) {

if (debugMode)
    serdb.printf("@ti_reinit.\n");

ti_init(myTz, useDLS);

} // ti_reinit

// ---------------------------------------------------------------------------
// Given timeStamp in GMT, this will set the MCU clock to the correct local time.
// It also returns the local time of the timeStamp

time_t ti_setRTC(
    time_t              timeStamp
    ) {

time_t                  oldTime;                            // current time of MCU's clock

// Setup 'tz' with correct timechangerule
Timezone                tzDls(tzRules[myTz][0],
                            tzRules[myTz][1]);
Timezone                tzNoDls(tzRules[myTz][1]);

if (debugMode)
    serdb.printf("@ti_setRTC: timeStamp: %lu\n", timeStamp);

// adjust for gmtOffset by multiplying hours to get seconds and adding to epock
if (useDLS)
    timeStamp = tzDls.toLocal(timeStamp);
else
    timeStamp = tzNoDls.toLocal(timeStamp);

// Set System Clock
oldTime = now();

#if defined rtcTeensy
    if (debugMode)
        serdb.printf("  Teensy internal RTC to %s", ctime(&timeStamp));
    Teensy3Clock.set(timeStamp);
    setTime(Teensy3Clock.get());
#elif defined rtcDS3231
    if (debugMode)
        serdb.printf("  Setting DS3231 to %s", ctime(&timeStamp));
    RTC.set(timeStamp);
    setTime(RTC.get());
#else // assume no RTC or rtcNone
    setTime(timeStamp);
    if (debugMode)
        serdb.printf("  MCU Clock set, but No RTC available!\n");
#endif

if (debugMode) {
    serdb.printf("  old time  : %02d/%02d/%04d %02d:%02d:%02d(%ld)\n", month(oldTime), day(oldTime), year(oldTime),
        hour(oldTime), minute(oldTime), second(oldTime), oldTime);
    serdb.printf("  new time  : %02d/%02d/%04d %02d:%02d:%02d(%ld)\n", month(timeStamp), day(timeStamp), year(timeStamp),
        hour(timeStamp), minute(timeStamp), second(timeStamp), timeStamp);
    serdb.printf("  correction: %d\n", (int)(timeStamp - oldTime));
    }

return timeStamp;

} // ti_setRTC

// ---------------------------------------------------------------------------
// if DLS time has changed, add or subtract an hour

// This function monitors for the time being changed due to DLS.

void ti_updDLS(
    ) {

static ulong            lockoutTimeout  = 1UL*1000UL;       // this timeout changes dynamically

static ulong            lastTime        = 0UL;              // what time was it, last time this func called
static ulong            lockoutTimer    = 0UL;

time_t                  oldTime;

// Setup 'tz' with correct timechangerule
Timezone                tzDls(tzRules[myTz][0],
                            tzRules[myTz][1]);
Timezone                tzNoDls(tzRules[myTz][1]);


//if (millis() % 1000 == 0) {
//    serdb.printf("@ti_updDLS: timpop: %d; useDLS: %d; ti_valid: %d\n",
//        timerPop(lockoutTimer, lockoutTimeout), useDLS, ti_valid);
//    delay(1);
//    }

// Are we not locked out, using DLS, and the time is valid?
if (timerPop(lockoutTimer, lockoutTimeout) && useDLS && ti_valid) {

    // The time can be adjusted by WWV, NTP, or ti_updDLS. If something else corrects
    // the time for DLS, ti_updDLS needs to NOT try to correct the new time. To do that
    // we look to see if the time has changed back or forward more than 55 minutes. If it
    // has, lockOutTimer is set to block ti_updDLS from attempting to make any changes for
    // the next 65 minutes.

    if (lastTime == 0UL)                                    // lastTime set? No, set it
        lastTime = now();
    else if (abs((int)(now() - lastTime)) > (55UL * 60UL)) {       // lastTime more than 55 minutes off? something has changed time.
        // lockout ti_updDLS from changing time
        if (debugMode)
            serdb.printf("@ti_updDLS: time changed manually. Locking out ti_updDLS\n");
        lockoutTimeout = 65UL * 60UL * 1000UL;
        lockoutTimer   = millis();
        lastTime       = 0UL;                               // Don't allow successive lockouts
        return;
        }
    else                                                    // nothing unusual - just update lastTime
        lastTime = now();

    if (ti_DLS != tzDls.locIsDST(now())) {                  // has time changed due to DLS?
        oldTime = now();
        if (ti_DLS)
            setTime(now()-(60*60));                         // fall back 60*60 seconds
        else
            setTime(now()+(60*60));

        #if defined rtcTeensy
            Teensy3Clock.set(now());
        #elif defined rtcDS3231
            RTC.set(now());
        #else // assume no RTC or rtcNone
            // there is no RTC to update with the new time
            if (debugMode)
                serdb.printf("  No RTC defined to be able to set!\n");
        #endif

        ti_DLS = !ti_DLS;

        lockoutTimeout = 65UL * 60UL * 1000UL;              // after DLS change don't check again for 65 minutes
        lockoutTimer = millis();                            // start lockout

        if (debugMode) {
            serdb.printf("@ti_updDLS: DLS has changed.\n");
            serdb.printf("  old Time: %s",   ctime(&oldTime));
            oldTime = now();
            serdb.printf("  new Time: %s",   ctime(&oldTime));
            serdb.printf("  ti_DLS:   %d\n", ti_DLS);
            }
        }
    else {                                                  // don't try again for a full minute
        lockoutTimeout = 1UL * 1000UL;                      // if DLS hasn't changed, timeout is 1 sec
        lockoutTimer = millis();
        }
    }

} // ti_updDLS
