// ----------------------------------------------------------------------------
// wwvb library

#pragma GCC diagnostic ignored "-Wwrite-strings"            // always ignore warning about using strings as constants
//#pragma GCC diagnostic ignored "-Wunused-variable"          // ignore unused variables until near end of development cycle, then comment this

#include                <Arduino.h>
#include                "globals.h"

#include                <time.h>                            // for ctime

// ---------------------------------------------------------------------------
// Public Globals
// external's in h file defined here
// ---------------------------------------------------------------------------

volatile uint16_t       wwv_abortCount  = 0;                // number of time wwv has given up trying to find valid packet
volatile float          wwv_corrAvg     = 0.0;              // # Average correction
volatile int            wwv_corrLast    = 0;                // # of seconds the clock was last corrected
volatile int            wwv_corrTotal   = 0;                // Total corrections
volatile wwv_decode_t   wwv_decode;                         // the decoded wwvb packet
volatile time_t         wwv_timeSet     = 0;                // time RTC was set by wwvb
volatile wwv_state_t    wwv_state       = wwv_invalid;      // wwv library state(listending, invalid, valid)
volatile uint16_t       wwv_syncCount   = 0;                // # of times RTC clock has been sync'ed to WWVB

// ---------------------------------------------------------------------------
// Private Globals
// ---------------------------------------------------------------------------

ulong                   attemptTimeout  = 0UL;              // How long to keep trying to connect in millisecs
ulong                   attemptTimer;                       // How long we've been trying to connect
IntervalTimer           radioTimer;                         // basic timer object to invoke wwv_tick every 10ms
int                     reqDuration;                        // requested number of hours to try to obtain signal
int                     reqStartHour;                       // requested hour we begin listening
bool                    active          = false;            // has wwv_init been called?

// ---------------------------------------------------------------------------
// Private Forward declarations

// ---------------------------------------------------------------------------
// Given timeStamp in GMT, this will set the MCU clock to the correct local time.
// It also returns the local time of the timeStamp

int getSignal(
    );

// ---------------------------------------------------------------------------
// Interrupt routine to rx and assemble the radio data

void wwv_tick(
    );

// ===========================================================================

// ---------------------------------------------------------------------------
// Read the radio pin, and return the NOT of the boolean value

int getSignal(
    ) {

return !digitalRead(pin_radioIn);                           // get current radio signal (reverse logic)

} // getSignal

//-----------------------------------------------------------------------------
// Disable wwv module. Removes the wwv_tick timer interrupt.

// calling wwv_end will set wwv_state to wwv_invalid since the rx packet is
// not valid.

void wwv_end(
    ) {

if (debugMode)
    serdb.printf("@wwv_end.\n");

radioTimer.end();

active     = false;
wwv_state  = wwv_invalid;

} // wwv_end


//-----------------------------------------------------------------------------
// initialize wwvlib module. Sets up the timer interrupt and enables it

/*
When wwv_init is called, it sets wwv_tick up to start immediately and run
until it gets a good signal.

startHour/duration are only used for successive signal searches. After the
signal is found, wwv_tick will stop until <startHour> (if startHour is 0,
wwv_tick will not restart until the next time it is 12:00AM).

Duration indicates the number of hours wwv_tick will try to receive a
packet. If a packet is rx'ed, it quits immediately. If a packet is never
rx'ed it will quit after <duration> hours have passed.

Duration may be set to 0 in which case after a good packet is rx'ed,
www_tick will reset and start looking for another good packet. When duration
is 0, <startHour> is ignored. Setting duration to 0 is typically used
just for debugging.

Since there is little possibility of rx'ing a good signal during daylight
hours, one would normally set startHour and duratin so that wwv_tick
runs at night. The default is start at 12:00AM and run for 6 hours which
is a reasonable bet.
*/

void wwv_init(
    int                 startHour,                          // starting hour
    int                 duration                            // duration in hours
    ) {

char                s[80];
char                timeStr[20];


if (debugMode) {
    snprintf(s, sizeof(s),
        "@wwv_init: %s; startHour: %d; duration: %d\n",
        ti_getDateTime(timeStr, sizeof(timeStr)), startHour, duration);
    serdb.printf("%s\n", s);
    }

reqStartHour = startHour;
reqDuration  = duration;

if (active) {
    if (debugMode)
        serdb.printf("  wwv_init already active!\n");
    // wwv_init can be called again to change startHour/duration, but the
    // existing timer is not changed
    return;
    }
else
    active = true;


// When wwv_init is first called, we will immediately start trying to get WWV time and
// will NOT timeout until it is rx'ed

attemptTimeout = 0UL;
attemptTimer   = millis();
wwv_state      = wwv_invalid;

radioTimer.begin(wwv_tick, 10000);                          // call wwv_tick every 10ms (10,000 microSecs)

} // wwv_init

//-----------------------------------------------------------------------------
// Interval interrupt function.

/*
The wwv_tick function is called every 10ms. It waits for a valid packet and then will
rx and assemble the radio data into <packet> and sets the RTC (real-time clock)
from that data.

This func will rx the WWVB signal without blocking. It does this by looking at the
signal every 10ms and determining the how long the signal is high. 0.2s is a '0',
0.5s is a '1' and 0.8s is a 'MARK'.

This function is implemented as a state machine with <state> being the current state.
When wwv_tick is called by the timer, it will execute the current state, then exit.
The state executed may or may not change the next state to be executed depending on
input.

<recentbits> is used to handle noise. When wwv_tick is reading a HIGH signal, it requires
a LOW signal be seen on 4 consecutive interrupts. Each signal is appended to the low
order bit of recentBits. If recentBits == 0, then the last 4 bits were all LOW and that
is considered the end of the HIGH signal. This allows occasional drops (errors) of
signal (zeros) in the data be read.
*/

void wwv_tick(
    ) {

const bool              dumpMark1       = false;            // if true disp getMark1 debugging
const bool              dumpOn          = false;            // if true, display additional debugging

enum                    bit_t {                             // type of bit rx'ed from wwvb
    low,                                                    // signal HIGH for 0.2 secs
    high,                                                   // signal HIGH for 0.5 secs
    mark,                                                   // signal HIGH for 0.8 secs
    };

enum                    state_t {                           // states for the wwv_tick state machine
    init_state,
    getMark1_state,
    getMark2_state,
    readPacketInit_state,
    readPacket_state,
    readPacketDone_state,
    decodeBits_state,
    done_state,
    timeout_state,
    waitForNextCycleInit_state,
    waitForNextCycle_state,
    };

// static variables' values are maintained between calls to the interrupt function

static uint16_t         lastSec         = 0;                // used to display current second being decoded.
static bit_t            packet[60];                         // bits of the packet being rx'ed - 60 bits, one for each second of signal
static uint8_t          packetIx;                           // next empty bit in packet array
static ulong            packetRxTimer;                      // time since packet started being read
static uint32_t         recentBits      = 0;                // last 4 states of the radioSignal while looking for header bits
static state_t          state           = init_state;       // current state of the state machine
static ulong            timer           = 0UL;              // length of time the signal has been in it's current hi/lo state
static time_t           timeStamp;                          // wwvb time in linux timeStamp format

// dynamic variables' values are lost between each call to the interrupt function

int                     day;
int                     month;
uint8_t                 radioSignal;                        // current radioSignal value (hi/lo)
char                    s[80];

switch (state) {

    case init_state:                                        // setup to start searching for 2  adjacent MARKs indicating beginning of transmission

        recentBits = 0;                                     // clear recentBits
        state      = getMark1_state;                        // proceed to searching for first MARK

        break;

    case getMark1_state:                                    // search for '1st' MARK from WWVB station

        if (attemptTimeout != 0UL && timerPop(attemptTimer, attemptTimeout)) {
            // we've run out of time w/o finding a good signal, so giveup
            state = timeout_state;
            break;
            }

        radioSignal = getSignal();

        if (radioSignal == LOW && recentBits == 0)          // If signal low and has been low, nothing to do
            break;

        if (radioSignal == HIGH && recentBits == 0) {       // beginning of a HIGH signal?
            if (dumpMark1)
                serdb.printf("1");
            recentBits = 1;                                 // beginning of HIGH, so set 1 recentBit high
            timer = millis() - 10;                          // init signal lenght timer. Subtract 10ms to account for first bit just read
            break;
            }

        if (dumpMark1) {
            serdb.printf("%d", radioSignal);
            if ((millis() / 1000UL) != lastSec) {
                lastSec = millis() / 1000;
                serdb.printf(" [%02ds] ", lastSec % 60);
                }
            }

        // If we get here, we are processing an existing HIGH signal

        recentBits = ((recentBits << 1) | radioSignal) & 0xF;

        if (recentBits == 0) {                              // last 4 signals rx'ed LOW?
            // 4 lows means 40ms of LOW signal, so process the last HIGH signal
            timer = millis() - timer - 40;                  // determine actual length of HIGH signal in millisecs
            if (timer >= 700 && timer <= 900) {             // did we RX a MARK?
                if (dumpOn)
                    serdb.printf("%s Marker found\n",
                        ti_getDateTime(s, sizeof(s)));
                state = getMark2_state;                     // we are ready to search for 2nd mark
                }
            else {
                if (dumpMark1)
                    serdb.printf(" ");
                state = init_state;                         // len of signal was incorrect for a mark so start over
                }
            break;
            }

        // if we get here, still processing a HIGH signal, so keep on doing that

        break;

    // Finding 2 MARKs in a row indicates beginning of packet

    case getMark2_state:                                    // 1st MARK seen, look if next signal rx'ed is MARK2

        if (attemptTimeout != 0UL && timerPop(attemptTimer, attemptTimeout)) {
            state = timeout_state;
            break;
            }

        radioSignal = getSignal();

        if (radioSignal == LOW && recentBits == 0)          // If signal low and has been low, nothing to do
            break;

        if (radioSignal == HIGH && recentBits == 0) {       // beginning of HIGH signal?
            if (dumpOn)
                serdb.printf("1");
            recentBits    = 1;                              // this is beginning, so set 1 recentBit high
            timer = millis() - 10;                          // init signal lenght timer. Subtract 10ms to account for first bit just read
            packetRxTimer = millis();                       // start timing time it takes to read entire packet
            break;
            }

        if (dumpOn) {
            serdb.printf("%d", radioSignal);
            if ((millis() / 1000UL) != lastSec) {
                lastSec = millis() / 1000;
                serdb.printf(" [%02ds] ", now() % 60);
                }
            }

        // If we get here, then we are processing an existing HIGH signal

        recentBits = ((recentBits << 1) | radioSignal) & 0xF;

        if (recentBits == 0) {                              // last 4 signal states low?
            // 4 lows mean 40ms of LOW signal, so process the last HIGHT signal
            timer = millis() - timer - 40;                  // determine actual legnth of HIGH signal in millisecs
            if (timer >= 700 && timer <= 900) {             // did we RX a MARK?
                if (debugMode)
                    serdb.printf("%s 2nd Marker found(%02ds)\n",
                        ti_getDateTime(s, sizeof(s)), now() % 60);
                state = readPacketInit_state;               // we've seen 2 marks, so start processing beginning of packet
                }
            else {
                if (dumpOn)
                    serdb.printf(" ");
                state = init_state;                         // len of signal was incorrect for a mark so start over
                }
            break;
            }

        // if we get here, still processing a HIGH signal, so keep on doing that

        break;

    // when MARK1 and MARK2 are rx'ed we have beginning of packet so setup
    // to begin reading the data into <packet>

    case readPacketInit_state:                              // setup to rx the packet

        if (debugMode)
            serdb.printf("@readPacketInit_state\n");

        packet[0] = mark;                                   // we've already read the first Mark
        packetIx  = 1;                                      // next bit read goes here
        timer     = millis() - 40;                          // when we started getting low signal  //***** I don't think I need this
        state     = readPacket_state;

        break;

    case readPacket_state:                                  // rx each bit until entire packet is built

        if (attemptTimeout != 0UL && timerPop(attemptTimer, attemptTimeout)) {
            state = timeout_state;
            break;
            }

        radioSignal = getSignal();

        if (radioSignal == LOW && recentBits == 0)          // wait for initial HIGH
            break;

        if (radioSignal == HIGH && recentBits == 0) {       // first bit of new signal
            if (dumpOn)
                serdb.printf("1");
            recentBits = 1;                                 // this is first HIGH bit, set recentBits accordingly
            timer = millis() - 10;                          // init signal lenght timer. Subtract 10ms to account for first bit just read
            break;
            }

        if (dumpOn) {
            serdb.printf("%d", radioSignal);
            if ((millis() / 1000UL) != lastSec) {
                lastSec = millis() / 1000;
                serdb.printf(" [%02ds] ", now() % 60);
                }
            }

        // If we get here, then we are in the middle of a HIGH signal, append the current signal state, and keep just the last 3 states
        recentBits = ((recentBits << 1) | radioSignal) & 0xF;

        if (recentBits == 0) {                              // if last 4 signals low, end of signal
            // decode the bit we rx'ed based on its length (of time)
            timer = millis() - timer - 40;                  // determine actual length of HIGH signal
            if (debugMode)
                serdb.printf("%03ld ", timer);
            if (timer >= 100 && timer <= 300)               // this range is from MAS6180C datasheet
                packet[packetIx] = low;
            else if (timer >= 400 && timer <= 600)          // this range is from MAS6180C datasheet
                packet[packetIx] = high;
            else if (timer >= 700 && timer <= 900)          // this range is from MAS6180C datasheet
                packet[packetIx] = mark;
            else {
                // weird signal length, start over
                if (dumpOn)
                    serdb.printf(" ");
                if (debugMode)
                    serdb.printf("@wwv_tick:readPacket_state: bad signal length: %ld. Restarting\n", timer);
                state = init_state;
                break;
                }

            packetIx = packetIx + 1;

            if (packetIx == 60) {                           // rx'ed entire packet?
                if (debugMode)
                    serdb.printf("\n");
                state = readPacketDone_state;
                break;
                }

            state    = readPacket_state;

            } // done rx'ing a bit

        break;

    // The whole packet has been read. Do some simple checks to verify it is correct

    case readPacketDone_state:                              // whole packet has been read

        if (debugMode) {
            serdb.printf("@readPacketDone_state:\n");

            for (packetIx = 0; packetIx < 60; packetIx = packetIx + 1) {
                switch (packet[packetIx]) {
                    case low:
                        serdb.printf("L");
                        break;
                    case high:
                        serdb.printf("H");
                        break;
                    case mark:
                        serdb.printf("M");
                        break;
                    } // switch
                } // for
                serdb.printf("\n");
            }

        // validate location of Marks

        if (    packet[9] != mark ||
                packet[19] != mark ||
                packet[29] != mark ||
                packet[39] != mark ||
                packet[49] != mark ||
                packet[59] != mark) {
            if (debugMode)
                serdb.printf("wwv_tick: missing Mark signal\n");
            state = init_state;
            break;
            }

        state = decodeBits_state;

        break;

    case decodeBits_state:                                  // decode the packet into the wwv_decode structure

        if (debugMode)
            serdb.printf("@decodeBits_state: %s\n", ti_getTime(s, sizeof(s)));

        // the method of decoding is brain-dead simple though  verbose. We look at every
        // bit in the packet. In many cases,  we simply add the BCD value of the bit to a running
        // tally of the associated value in wwv_decode.

        for (packetIx = 0; packetIx <= 58; packetIx = packetIx + 1) {
            switch (packetIx) {
                case 0:                                     // Mark ignore
                    break;
                case 1:                                     // minute: 40's
                    wwv_decode.minutes = 0;                 // initialize
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 40;
                    break;
                case 2:                                     // minute: 20's
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 20;
                    break;
                case 3:                                     // minute: 10's
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 10;
                    break;
                case 4:                                     // unused
                    break;
                case 5:                                     // minutes: 8
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 8;
                    break;
                case 6:                                     // minutes: 4
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 4;
                    break;
                case 7:                                     // minutes: 2
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 2;
                    break;
                case 8:                                     // minutes: 1
                    if (packet[packetIx] == high)
                        wwv_decode.minutes = wwv_decode.minutes + 1;
                    break;
                case 9:                                     // Mark ignore
                    break;
                case 10:                                    // unused
                    break;
                case 11:                                    // unused
                    break;
                case 12:                                    // hours: 20
                    wwv_decode.hours = 0;                   // initialize
                    if (packet[packetIx] == high)
                        wwv_decode.hours = wwv_decode.hours + 20;
                    break;
                case 13:                                    // hours: 10
                    if (packet[packetIx] == high)
                        wwv_decode.hours = wwv_decode.hours + 10;
                    break;
                case 14:                                    // unused
                    break;
                case 15:                                    // hours: 8
                    if (packet[packetIx] == high)
                        wwv_decode.hours = wwv_decode.hours + 8;
                    break;
                case 16:                                    // hours: 4
                    if (packet[packetIx] == high)
                        wwv_decode.hours = wwv_decode.hours + 4;
                    break;
                case 17:                                    // hours: 2
                    if (packet[packetIx] == high)
                        wwv_decode.hours = wwv_decode.hours + 2;
                    break;
                case 18:                                    // hours: 1
                    if (packet[packetIx] == high)
                        wwv_decode.hours = wwv_decode.hours + 1;
                    break;
                case 19:                                    // Mark ignore
                    break;
                case 20:                                    // unused
                    break;
                case 21:                                    // unused
                    break;
                case 22:                                    // days: 200
                    wwv_decode.dayOfYear = 0;               // initialize
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 200;
                    break;
                case 23:                                    // days: 100
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 100;
                    break;
                case 24:                                    // unused
                    break;
                case 25:                                    // days: 80
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 80;
                    break;
                case 26:                                    // days: 40
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 40;
                    break;
                case 27:                                    // days: 20
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 20;
                    break;
                case 28:                                    // days: 10
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 10;
                    break;
                case 29:                                    // Mark ignore
                    break;
                case 30:                                    // days: 8
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 8;
                    break;
                case 31:                                    // days: 4
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 4;
                    break;
                case 32:                                    // days: 2
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 2;
                    break;
                case 33:                                    // days: 1
                    if (packet[packetIx] == high)
                        wwv_decode.dayOfYear = wwv_decode.dayOfYear + 1;
                    break;
                case 34:                                    // unused
                    break;
                case 35:                                    // unused
                    break;
                case 36:                                    // DUT1 sign
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetSign = 0b100;
                    else
                        wwv_decode.dut1OffsetSign = 0;
                    break;
                case 37:                                    // DUT1 sign
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetSign = wwv_decode.dut1OffsetSign | 0b010;
                    break;
                case 38:                                    // DUT1 sign
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetSign = wwv_decode.dut1OffsetSign | 0b001;
                    break;
                case 39:                                    // Mark ignore
                    break;
                case 40:                                    // DUT Value 0.8
                    wwv_decode.dut1OffsetValue = 0.0;
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetValue = wwv_decode.dut1OffsetValue + 0.8;
                case 41:                                    // DUT Value 0.4
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetValue = wwv_decode.dut1OffsetValue + 0.4;
                case 42:                                    // DUT Value 0.2
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetValue = wwv_decode.dut1OffsetValue + 0.2;
                case 43:                                    // DUT Value 0.1
                    if (packet[packetIx] == high)
                        wwv_decode.dut1OffsetValue = wwv_decode.dut1OffsetValue + 0.1;
                case 44:                                    // unused
                    break;
                case 45:                                    // year: 80
                    wwv_decode.year = 2000;
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 80;
                    break;
                case 46:                                    // year: 40
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 40;
                    break;
                case 47:                                    // year: 20
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 20;
                    break;
                case 48:                                    // year: 10
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 10;
                    break;
                case 49:                                    // Mark ignore
                    break;
                case 50:                                    // year: 8
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 8;
                    break;
                case 51:                                    // year: 4
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 4;
                    break;
                case 52:                                    // year: 2
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 2;
                    break;
                case 53:                                    // year: 1
                    if (packet[packetIx] == high)
                        wwv_decode.year = wwv_decode.year + 1;
                    break;
                case 54:                                    // unused
                    break;
                case 55:                                    // leap year indicator
                    if (packet[packetIx] == high)
                        wwv_decode.leapYear = 1;
                    else
                        wwv_decode.leapYear = 0;
                    break;
                case 56:                                    // Leap Second
                    if (packet[packetIx] == high)
                        wwv_decode.leapSecond = 1;
                    else
                        wwv_decode.leapSecond = 0;
                    break;
                case 57:                                    // DST Indicator
                    if (packet[packetIx] == high)
                        wwv_decode.dstIndicator = 0b10;
                    else
                        wwv_decode.dstIndicator = 0;
                    break;
                case 58:                                    // DST Indicator
                    if (packet[packetIx] == high)
                        wwv_decode.dstIndicator = wwv_decode.dstIndicator | 0b01;
                    break;

                default:
                    break;

                } // switch
            } // for

        // verify all decoded values are within range

        if (wwv_decode.year < 2023 || wwv_decode.year > 2199) {
            state = init_state;
            break;
            }
        if (wwv_decode.dayOfYear < 1 || wwv_decode.dayOfYear > 366) {
            state = init_state;
            break;
            }
        if (wwv_decode.hours < 0 || wwv_decode.hours > 23) {
            state = init_state;
            break;
            }
        if (wwv_decode.minutes < 0 || wwv_decode.minutes > 59) {
            state = init_state;
            break;
            }

        // determine mm/dd using dayOfYear
        dayOfYearToDate(wwv_decode.year, wwv_decode.dayOfYear, &month, &day);

        // Convert yymmdd hhmm to linux timestamp
        timeStamp = dateTime2timeStamp(wwv_decode.year, month, day, wwv_decode.hours,
            wwv_decode.minutes, 0);

        // compute time(in millisecs) since we rx'ed the beginning of the time packet
        packetRxTimer = millis() - packetRxTimer;

        // If packetRxTimer is not close to 60 seconds, something went wrong
        if (packetRxTimer < 59500 || packetRxTimer > 60000) {
            if (debugMode)
                serdb.printf("packetRxTimer out of range: %lu\n", packetRxTimer);
            state = init_state;
            break;
            }

        // The time stamp we have rx'ed is exactly 1 minute AFTER we will set the time, so adj timeStamp
        timeStamp = timeStamp + 60;

        // packetRxTimer will be a little < than 60 seconds, so wait until exactly 60 seconds have
        // passed, then set the time
        if (packetRxTimer < 60000)
            delay(60000 - packetRxTimer);
        wwv_corrLast  = now();                              // need to know current time to compute correction
        wwv_timeSet   = ti_setRTC(timeStamp);               // set new time in RTC
        wwv_corrLast  = wwv_timeSet - wwv_corrLast;         // correction in seconds (+ means we added seconds)
        if (abs(wwv_corrLast) > 30) {
            // The correction is too big to just be RTC drifting.  Most likely the clock has been manually set/reset
            // In this case we want to increment wwv_syncCount, but not mess up wwv_corr_avg
            wwv_syncCount = wwv_syncCount + 1;
            wwv_corrTotal = wwv_corrTotal + round(wwv_corrAvg);
            wwv_corrAvg   = (float)wwv_corrTotal / wwv_syncCount;
            }
        else {
            // the correction is small in which case we want to add compute wwv_corrAvg properly.
            wwv_syncCount = wwv_syncCount + 1;              // inc # of times we've synced to WWVB
            wwv_corrTotal = wwv_corrTotal + wwv_corrLast;   // running summary of total time corrections
            wwv_corrAvg   = (float)wwv_corrTotal / wwv_syncCount; // get avg correction.
            }

        // do this AFTER setRTC as it may delay code too long
        if (true/*dumpOn*/) {
            serdb.printf("%s: decoded wwvb packet: %03d/%02d %02d:%02d:00 GMT\n",
                ti_getDateTime(s,sizeof(s)),
                wwv_decode.dayOfYear, wwv_decode.year,
                wwv_decode.hours, wwv_decode.minutes);
            serdb.printf("  DUT1 sign: %X; Offset: %f; Leap Year: %d; Leap Second: %d; DST: %X\n",
                wwv_decode.dut1OffsetSign, wwv_decode.dut1OffsetValue,
                wwv_decode.leapYear, wwv_decode.leapSecond, wwv_decode.dstIndicator);
            serdb.printf("  mo: %d; da: %d\n", month, day);
            serdb.printf("  timeStamp    : %lu\n", timeStamp);
            serdb.printf("  wwv_corrLast : %d\n", wwv_corrLast);
            serdb.printf("  wwv_syncCount: %d\n", wwv_syncCount);
            serdb.printf("  wwv_corrTotal: %d\n", wwv_corrTotal);
            serdb.printf("  wwv_corrAvg  : %f\n", wwv_corrAvg);
            serdb.printf("  packetRxTimer: %d\n", packetRxTimer);
            }

        state = done_state;

        break;

    case done_state:                                        // done processing the rx'ed packet

        // if time has been invalid we RE-initialize ti_init
        // to make sure the time module is happy

        if (!ti_valid)
            ti_reinit();

        wwv_state  = wwv_valid;
        state      = waitForNextCycleInit_state;

        break;

    case timeout_state:                                     // cleanup if we timeout while rx'ing signal

        wwv_state      = wwv_invalid;
        wwv_abortCount = wwv_abortCount + 1;

        state          = waitForNextCycleInit_state;

        break;

    case waitForNextCycleInit_state:                        // initialize wait for next cycle

        if (debugMode)
            serdb.printf("@waitForNextCycleInit_state: reqStartHour: %d\n", reqStartHour);

        // We are here because we completed a good cycle.

        // if requested duration is 0, that means we start over immediately, never to stop

        if (reqDuration == 0) {
            if (debugMode)
                serdb.printf("  duration is 0, restart immediately.\n");
            attemptTimeout = 0UL;
            state = init_state;
            break;
            }

        //Setup to wait until <reqHourStart>

        if (hour(now()) == reqStartHour) {                  // is it already <reqHourStart>?
            // do not allow an immdiate restart - start checking again in 1 hour
            attemptTimeout = 1UL * 60UL * 60UL * 1000UL;    // wait 1 hour before attempting restart
            attemptTimer   = millis();
            }
        else {
            // make timer in waitForNextCycle_state pop immediately
            attemptTimeout = 0UL;
            attemptTimer   = millis();
            }

        state = waitForNextCycle_state;

        break;

    // This state will wait until it is time to start looking for a wwv signal again.

    case waitForNextCycle_state:                            // wait until next cycle should begin

        if (timerPop(attemptTimer, attemptTimeout)) {
            if (hour(now()) == reqStartHour) {              // is it time to start listening?
                // We've rx'ed a good time at least once, so
                // setup attemptTimer to give up if duration is exceeded
                attemptTimeout = reqDuration * 60UL * 60UL * 1000;
                attemptTimer   = millis();
                state = init_state;
                }
            else {
                // not time to start, we'll check again in 1 minute
                attemptTimeout = 60UL * 1000UL;             // ck in 1 minute
                attemptTimer   = millis();
                }
            }

        break;

    default:                                                // we should never get here

        if (debugMode)
            serdb.printf("@wwv_tick: default case hit!!\n");

        state = init_state;

        break;

    } // switch

} // wwv_tick

// ---------------------------------------------------------------------------
// return strength of the wwvb signal, 0-10.

/*

wwv_strength is useful to align the antenna for the best signal possible.
If you are not receiving a strength of 10, you will not rx a valid packet.

The actual "strength" cannot be determined by looking at the wwvb break out
board. It simply returns HIGH/LOW depending on how the radio signal is modulated.
So, instead I look at the signal to see how 'noisy' the signal is.

Each second, only 1 'bit' is sent. Regardless of the length of the bit there will be
1 transition to high and 1 transition to low.

So if I monitor for x seconds, there should be x transitions to high.

I will be sampling 100 times / second. If I see 1 transition per second, that
is a perfect signal or '10'. If I get 50 transitions in 100 samples,
that would be the worst possible signal or '0'.
If I get no transitions, that is no signal which would also be a '0'.

After experimenting with this algorithm, I found a linear conversion
doesn't really work. If 1 transition is a STR(strength) of 10 and 50
transitions is STR of 0, 25 transitions giving a strength of 5 doesn't
help much because in a very noisy and unusable environment I'm getting
< 10 transitions / sec.

To make the STR value more usable, I use a LOG scale so the upper end
is more sensitive. Table of some values:

Transitions/Sec     Strength
2                   9
3                   9
5                   8
10                  6
20                  3
40                  1

I have not been able to conclusively prove wwv_sigStrenth doesn't impact wwv_tick.
It shouldn't and it doesn't when testing with emulator. But in actual operating
conditions, it seems to sync less often than I would expect.

*/

int wwv_sigStrength(
    ) {

const int               seconds         = 3;                // number of seconds to sample
const int               frequency       = 100;              // samples per second

ulong                   duration        = 1000 / frequency; // millisecs per sample
float                   f;                                  // temp float
uint8_t                 lastSignal;                         // prior radioSignal
uint8_t                 radioSignal;                        // current radioSignal value (hi/lo)
char                    s[40];
int                     sample;                             // sample count within each second (0 thru <frequency>-1)
int                     secondNo;                           // current sample second (0 to sampleLen-1)
int                     strength;
int                     transitions;                        // number of transitions during all samples

if (debugMode)
    serdb.printf("@wwv_sigStrength (%s): ", ti_getTime(s, sizeof(s)));

lastSignal  = !digitalRead(pin_radioIn);                     // prime the pump
transitions = 0;

for (secondNo = 0; secondNo < seconds; secondNo = secondNo + 1) {
    for (sample = 0; sample < frequency; sample = sample + 1) {
        if (serdb.available()) {
            return 0;
            }
        radioSignal = getSignal();
        if (lastSignal == LOW && radioSignal == HIGH) {
            // we transitions from low to high, so count it
            transitions = transitions + 1;
            }
        lastSignal = radioSignal;
        delay(duration);
        } // for - sample
    } // for - second

if (debugMode)
    serdb.printf("Trans: %d; ", transitions);

transitions = round((float)transitions/seconds);

if (debugMode)
    serdb.printf("T/sec: %d; ", transitions);

if (transitions == 1) {
    // 1T/sec is perfect
    strength = 10;
    }
else if (transitions == 0) {
    // 0T/sec is no signal at all
    strength = 0;
    }
else {
    f = 1-((transitions - 2)/48.0);                         // Convert transitions from 2-50 to 0-1 scale (0 being worst)
    strength = round(pow(10, f)) - 1;                              // Strength is 10**f-1 which is log scale of 0 - 9 (0 worst)
    if (debugMode)
        serdb.printf("f: %.3f; ", f);
    }

if (debugMode)
    serdb.printf("Strength: %d\n", strength);

return strength;

} // wwv_sigStrength

